from TDhelper.generic.transformationType import transformation
from TDhelper.document.excel.FieldType import *
from openpyxl import load_workbook
import datetime
import os
import csv
from types import FunctionType,MethodType





class Meta:
    file = None
    sheet = "sheet1"
    extension = "xlsx"
    headrow = True 
    headrow_row_offset = 1

class modelMeta(type):
    def __new__(cls, name, bases, dct):
        attrs = {
            "mapping": {},
            "__reverse_mapping__": {},
            "__enter__": __enter__,
            "__exit__": __exit__,
            "__rows__": [],
            "close": close,
            "__initExcelHandle__": __initExcelHandle__,
            "_translateMode": _translateMode,
            "__rowsCount__": 0,
            "__col_offset__": 0, 
        }
        for name, value in dct.items():
            if isinstance(dct[name], FieldType):
                attrs["mapping"][name] = value.bindCol
                attrs["__reverse_mapping__"][str(value.bindCol)] = name
                attrs[name] = cls._AttributeOverride(name, value.fieldType)
            elif isinstance(dct[name],FunctionType):
                _n_=name.lower()
                if _n_ == "getcount":
                    attrs[name]= getCount
                elif _n_ == "items":
                    attrs[name]= items
                elif _n_=="close":
                    attrs[name]= close
                elif _n_=="range_items":
                    attrs[name]= items_range
                elif _n_=="getitem":
                    attrs[name]= getItem
                elif _n_=="save_excel":
                    attrs[name]= save_excel
                else:
                    attrs[name]= value
            else:
                attrs[name] = value
        return super(modelMeta, cls).__new__(cls, name, bases, attrs)

    class _AttributeOverride:
        def __init__(self, name, m_type):
            self._name = name
            self._type = m_type

        def __get__(self, instance, owen):
            return instance.__dict__[self._name]

        def __set__(self, instance, value):
            try:
                instance.__dict__[self._name] = transformation(value, self._type)
                instance.__rows__[instance.__col_offset__-1][instance.mapping[self._name]-1].value= value
            except Exception as e:
                raise e
            
        def __delete__(self, instance):
            instance.__dict__.pop(self._name)

def __initExcelHandle__(self):
    try:
        if self.Meta.file:
            m_extension = self.Meta.file.rsplit(".")[1]
            if m_extension == "csv":
                self.Meta.extension = "csv"
                self.__excelHandle__ = open(self.Meta.file)
                self.__sheetHandle__ = csv.reader(self.__excelHandle__)
            elif m_extension == "xlsx" or m_extension == "xls":
                self.Meta.extension = "xlsx"
                self.__excelHandle__ = load_workbook(self.Meta.file)
                self.__sheetHandle__ = self.__excelHandle__[self.Meta.sheet]
            else:
                raise Exception("file extension is error.")
            if self.__sheetHandle__:
                for col in self.__sheetHandle__:
                    self.__rows__.append(col)
                if self.Meta.headrow:
                    for _t_ in range(0,self.Meta.headrow_row_offset):
                        self.__rows__.pop(0)
                self.__rowsCount__ = len(self.__rows__)
        else:
            raise Exception("meta file is None.")
    except Exception as e:
        raise e


def _translateMode(self, data=[]):
    if data:
        try:
            tmp_index = 1
            for v in data:
                if v == None or v == "None" or v=="#N/A":
                    v = None
                if str(tmp_index) in self.__reverse_mapping__:
                    setattr(
                        self,
                        self.__reverse_mapping__[str(tmp_index)],
                        v if not hasattr(v, "value") else v.value,
                    )
                tmp_index += 1
            return self
        except Exception as e:
            raise e
    else:
        raise Exception("translate data is none.")


def items(self):
    assert self.Meta.file, "Meta file is none."
    self.__col_offset__=0
    for col in self.__rows__:
        self.__col_offset__+=1
        yield self._translateMode(col)

def items_range(self,star:int=0,end:int=-1):
    if star<0:
        raise Exception("params 'star' value error.")
    if end==-1:
        end= len(self.__rows__)
    elif end >= len(self.__rows__):
        raise Exception("params 'end' out of range.")
    self.__col_offset__= star
    for o in range(star,end):
        result = self._translateMode(self.__rows__[o])
        self.__col_offset__+=1
        yield result
        
def getItem(self, offset):
    assert 0 <= offset < self.__rowsCount__, "offset %d, " % offset
    self.__col_offset__= offset
    return self._translateMode(self.__rows__[offset])

def __enter__(self):
    return self

def __exit__(self, exc_type, exc_value, exc_t):
    self.close()

def save_excel(self,save_path:str,auto_name=True):
    if os.path.exists(save_path):
        if not auto_name:
            raise Exception("%s file has exists." % save_path) 
        else:
            save_path = save_path.replace(".csv",datetime.datetime.now().strftime("%Y_%m_%d %H_%m_%S")+".csv")
    try:
        with open(save_path,"a+",766) as f:
            for col in self.__rows__:
                row_str= ",".join([o.value for o in col])
                row_str=",".join([row_str,"\r"])
                f.write(row_str)
                f.flush()
            f.close()
        return True
    except Exception as e:
        raise e
    
def close(self):
    self.__excelHandle__.close()
    if self.Meta.extension.lower() != "csv":
        pass#self.__sheetHandle__.close()
    self.__excelHandle__ = None
    self.__sheetHandle__ = None
    self.Meta = None


def getCount(self):
    return self.__rowsCount__



from TDhelper.cache.webCache.interface import cacheInterface
from TDhelper.db.mongodb.setting import db_cfg
from TDhelper.db.mongodb.base import mongodbclient
from TDhelper.db.mongodb.dbhelper import dbhelper
from TDhelper.db.mongodb.objectId import objectId

class mongo(cacheInterface):
    def __init__(self,**kwargs) -> None:
        assert 'options' in kwargs, "not found 'options' in conf"
        assert 'host' in kwargs['options'], "not found 'host' in conf options node."
        assert 'db' in kwargs['options'], "not found 'db' in conf options node."
        db_cfg['url']=kwargs['options']['host']
        db_cfg['db']=kwargs['options']['db']
        db_cfg['user']=kwargs['options']['user'] if 'user' in kwargs['options'] else None
        db_cfg['passwd']=kwargs['options']['passwd'] if 'passwd' in kwargs['options'] else None
        super(mongo,self).__init__()
    
    def collect(self,k):
        assert k in self.__cache__, "can not found collect '%s'"% k
        self.__cursor__= self.__cache__[k]
        return self
    
    def addCollect(self,k,v):
        if k in self.__cache__:
            raise Exception("'%s' has already in cache."% k)
        else:
            self.__cache__[k]= type(v.__name__,(v,),{})()
            return self
            
    def delCollect(self,k):
        if k in self.__cache__:
            del self.__cache__[k]
            return self
        else:
            raise Exception("'%s' not found in cache."%k)
    
    def set(self,*args,**kwargs):
        if self.__cursor__:
            self.__cursor__.save(kwargs)
            self.__cursor__=None
            return kwargs
        else:
            raise Exception('not found collect.')
        
    def exist(self,*args,**kwargs):
        if self.__cursor__:
            flag= False
            if self.__cursor__.findOne(kwargs):
                flag= True
            self.__cursor__= None
            return flag
        else:
            raise Exception("not found collect.")
        
    def get(self,flag="single",*args,**kwargs)-> object:
        if self.__cursor__:
            result= None
            if flag.lower()=="single":
                result= self.__cursor__.findOne(kwargs)
            else:
                result= self.__cursor__.find(**kwargs)
            self.__cursor__= None
            return result
        else:
            raise Exception("not found collect.")
    
    def remove(self,*args,**kwargs):
        if self.__cursor__:
            result= self.__cursor__.remove(kwargs)
            self.__cursor__= None
            return result
        else:
            raise Exception('not found collect.')
    
    def update(self,*args,**kwargs):
        if self.__cursor__:
            assert args.__len__()>0, "not found query argument."
            result= self.__cursor__.update(args[0],kwargs)
            self.__cursor__= None
            return result
        else:
            raise Exception('not found collect.')


from TDhelper.cache.webCache.mongo import mongo
class webCacheFactory:
    __cache__={}
    __cursor__=None
    __handle__=None
    def __init__(self,conf={}):
        assert 'ENGINE' in conf, "'ENGINE' not in conf"
        cls= conf['ENGINE'].lower()
        if cls=="memory":
            pass
        elif cls=="mongo":
            self.__handle__= mongo(**conf)
        elif cls=="redis":
            pass
        else:
            raise Exception("cls must value: memory | mongo | redis.")
    
    def set(self,*args,**kwargs):
        return self.__handle__.set(*args,**kwargs)
        
    def get(self,flag='single',*args,**kwargs):
        return self.__handle__.get(flag,*args,**kwargs)

    def exist(self,*args,**kwargs):
        return self.__handle__.exist(*args,**kwargs)
    
    def collect(self,k):
        self.__handle__.collect(k)
        return self
    
    def addCollect(self,k,v):
        return self.__handle__.addCollect(k,v)
         
    def delCollect(self,k):
        return self.__handle__.delCollect(k)
    
    def remove(self,*args,**kwargs):
        return self.__handle__.remove(*args,**kwargs)
    
    def update(self,*args,**kwargs):
        return self.__handle__.update(*args,**kwargs)


from .meta.modelMeta import modelMeta

class model(metaclass=modelMeta):
    __excelHandle__ = None  # excel object.
    __sheetHandle__ = None  # sheet object.
    __rows__ = []  # all rows.

    def __init__(self, excelPath:str, headrow:bool=True, headrow_rows_count:int=1):
        '''初始化
        params:   
            - excelPath: <string>, Excel文件路径
            - headrow: <bool>, 是否有表头, 默认true
            - headrow_rows_count: <int>, 表头有多少行， 默认1
        
        '''
        assert excelPath, "parameter <excelPath> is None."
        self.Meta.file = excelPath
        self.Meta.headrow = headrow
        self.Meta.headrow_row_offset = headrow_rows_count
        try:
            self.__initExcelHandle__()
        except Exception as e:
            raise e

    def __initExcelHandle__(self):
        return None
    
    def items(self):
        pass

    def items_range(self,star:int=0,end:int=-1):
        pass
            
    def getItem(self, offset):
        pass

    def save_excel(self,save_path:str,auto_name=True):
        pass
        
    def close(self):
        pass

    def getCount(self):
        pass

    class Meta:
        """
        元数据
        - file: <string>, 文件路径.
        - sheet: <string>, sheet名称.
        - extension: <string>, 文件类型(xlsx,csv)
        """
        file = ""
        sheet = "sheet1"
        extension = "xlsx"
        headrow = True
        headrow_row_offset = 1



from copy import deepcopy
import random
import time
import logging
from TDhelper.network.http.REST_HTTP import GET
from threading import Thread

class HostManager:
    def __init__(self) -> None:
        self.__host_cnf_backup__ = {}
        self.__host_cnf__ = {}
        self.__host__ = []
        self.__host_len__ = 0
        self.__monitor_speed__= 10
        
        # self.__health_sniffer__ = ""
        self.__monitor_state__= True
    def __backup_host__(self):
        self.__host_cnf_backup__ = deepcopy(self.__host_cnf__)

    def get_per_http_host(self):
        try:
            return self.__generateHost__(self.getHost())
        except Exception as e:
            raise e

    def getHost(self,exclude:list=[]):
        """getHost
            get an access host cnf.
        Parameters:
            self - <class: HostManager>, hostManager instance.
        Returns:
            success - <class, json>, host cnf json.
            fail - None.
        Raises:
            None
        """
        result = None
        retry = 3
        while (True and retry > 0):
            v = 0 if self.__host_len__ == 1 else random.randint(
                0, self.__host_len__-1)
            if exclude:
                if exclude.__contains__(v):
                    continue
            result = self.__host_cnf__[self.__host__[v]]
            result["serverId"] = self.__host__[v]
            if result['status']:
                break
            retry -= 1
        return result

    def del_host(self,k):
        pass
    
    def get_host_by_key(self, key):
        return self.__host_cnf__[key]

    def register(self, serverId, host, port: int = 80, sniffer="/api/sniffer", proto: str = "http://", status=True):
        """register
            register an rpc host
        Parameters:
            self - <class: HostManager>, hostManager instance.
            serverId - <class: int>, serverId
            host - <class: str>, ip or domain
            port - <class: int>, port. default 80
            sniffer - <class: str>, the healthy sniffer uri. default /api/sniffer
            proto - <class: str>, http:// or https://. default http://
            status - <class: bool>, default True
        Returns:
            None
        Raises:
            10000 - key already exists.
        """
        serverId = str(serverId)
        proto = self.__checken_proto__(proto)
        sniffer= sniffer if sniffer else "/api/sniffer"
        if serverId not in self.__host_cnf__:
            self.__host_cnf__[serverId] = {
                "proto": proto,
                "host": host,
                "port": port,
                "status": status,
                "sniffer": sniffer
            }
            if serverId not in self.__host_cnf_backup__:
                self.__host_cnf_backup__ = {
                    "proto": proto,
                    "host": host,
                    "port": port,
                    "status": status,
                    "sniffer": sniffer
                }
            self.__host__.append(serverId)
            self.__host_len__ = len(self.__host__)
        else:
            raise (Exception("10000, '%s' already exists." % serverId))

    def __checken_proto__(self, v):
        v = v.replace(":", "")
        v = v.replace("/", "")
        v = v+"://"
        return v

    def reload(self):
        '''reload
            reload host conf by host conf backup.
        '''
        self.__host_cnf__ = {}
        self.__host__ = []
        self.__host_len__ = 0
        # self.__health_sniffer__ = ""
        for k, v in self.__host_cnf_backup__.items():
            self.register(k, v["host"], v["port"],
                          v["sniffer"], v["proto"], v["status"])

    def delete(self, serverId):
        """delete
            delete a host cnf.
        Parameters:
            self - <class, HostManager>, hostmanager instance.
            serverId - <class, int|str>, serverId.
        Returns:
            success - serverId
            fail - None
        Raises:
            None.
        """
        if str(serverId) in self.__host_cnf__:
            m_result = self.__host_cnf__.pop(serverId)
            self.__host__.remove(serverId)
            self.__host_len__ = len(self.__host__)
            return m_result
        else:
            return None

    def health(self):
        """health
            check host health status.
        Parameters:
            self - <class, HostManager>, hostmanager instance.
        Returns:
            <class, json>, cheked result.
            json desc
                statistics - <class, json>, checked result.
                    total - host count.
                    health - checked it's success count.
                    unhealthy - checked it's fail count.
                health - <class, list>, health serverId list.
                unhealthy - <class, list>, unhealthy serverId list.
        Raises:
            None.
        """
        while self.__monitor_state__:
            result = {
                "statistics": {
                    "total": self.__host_len__,
                    "health": 0,
                    "unhealthy": 0,
                },
                "health": [],
                "unhealthy": []
            }
            for k in self.__host__:
                m_uri = self.__generateHost__(self.__host_cnf__[k])
                if m_uri:
                    m_uri += "/" + \
                        self.__host_cnf__[k]["sniffer"].strip(
                            '/').replace('\\', "/")
                    status, body = GET(m_uri, time_out=5)
                    if status == 200:
                        self.__host_cnf__[k]['status'] = True
                        logging.info("%s%s:%s%s  success."%(self.__host_cnf__[k]['proto'],self.__host_cnf__[k]['host'],self.__host_cnf__[k]['port'],self.__host_cnf__[k]['sniffer']))
                        result["health"].append(k)
                    else:
                        self.__host_cnf__[k]["status"] = False
                        logging.info("%s%s:%s%s fail."%(self.__host_cnf__[k]['proto'],self.__host_cnf__[k]['host'],self.__host_cnf__[k]['port'],self.__host_cnf__[k]['sniffer']))
                        result["unhealthy"].append(k)
                else:
                    self.__host_cnf__[k]["status"] = False
                    logging.info("%s%s:%s%s fail."%(self.__host_cnf__[k]['proto'],self.__host_cnf__[k]['host'],self.__host_cnf__[k]['port'],self.__host_cnf__[k]['sniffer']))
                    result['unhealthy'].append(k)
            for k in result["unhealthy"]:
                self.delete(k)
            result["statistics"]["health"] = len(result["health"])
            result["statistics"]["unhealthy"] = len(result["unhealthy"])
            time.sleep(self.__monitor_speed__)

    def __start_state_monitor__(self, speed= None):
        self.__monitor_state__= True
        self.__monitor_speed__= speed if speed else self.__monitor_speed__
        monitor= Thread(target= self.health)
        monitor.start()
        
    def __stop_state_monitor__(self):
        self.__monitor_state__= False
        self.__monitor_speed__= 10
        
    def __generateHost__(self, uri_cnf: dict):
        proto = uri_cnf["proto"]
        host = uri_cnf["host"]
        port = str(uri_cnf["port"])
        if uri_cnf["status"]:
            if port != 80 or port != "80":
                return proto+host+":"+port
            else:
                return proto+host
        else:
            return None



from enum import Enum
import enum
import hashlib
import copy

class Protocol:
    def __init__(self, protocolver='td-protocol1.0', charset='utf-8'):
        self.ver = protocolver  # 协议版本
        self.index = -1  # 包序号,分包需要添加一个总包数；在业务逻辑上服务端、客户端做断点续传应答（模仿迅雷，BT加临时文件解决）
        self.charset = charset  # 数据块编码
        self.dataLength = 0  # 数据块长度
        self.dataMd5 = None  # 数据哈希
        self.command= b'' #命令区域
        self.data = b''  # 数据

    def decode(self)->bytes:
        '''
            将封装好的协议包转为bytes数据准备发送
        '''
        try:
            m_str:bytes = b''
            for item in self.__dict__:
                if isinstance(self.__dict__[item],bytes):
                    m_str+= self.__dict__[item]
                elif isinstance(self.__dict__[item],int):
                    m_str+= self.__dict__[item].to_bytes(4,byteorder='little',signed=False)
                else:
                    m_str += bytes(self.__dict__[item],encoding=self.charset)
                m_str+= bytes("\r",encoding=self.charset)
            pass
        except Exception as e:
            raise e
        return m_str
    
    def setcommand(self,cmd:str):
        self.command=bytes(cmd,encoding=self.charset)
        
    def getcommand(self):
        return self.command
        
    def encode(self, buff: bytes | str, index=1):
        '''
            对数据进行协议封装
        '''
        self.data = buff if isinstance(buff,bytes) else bytes(buff,encoding=self.charset)
        m_md5 = hashlib.md5()
        m_md5.update(self.data)
        self.dataMd5 = m_md5.hexdigest()
        self.dataLength = len(self.data)
        self.index = index

    def checkMD5(self):
        m_md5 = hashlib.md5()
        m_md5.update(self.data.encode(encoding=self.charset))
        if self.dataMd5 == m_md5.hexdigest():
            return True
        else:
            return False

    def clear(self):
        self.index = -1  # 包序号
        #self.charset = ''  # 数据块编码
        self.dataLength = 0  # 数据块长度
        self.dataMd5 = None  # 数据哈希
        self.data = b''  # 数据


class analysis:
    def __init__(self, protocol: Protocol):
        self._protocol = protocol  # 协议
        self._buff = b''  # 接受缓存
        self._buffOffset = 0  # 数据缓存游标
        self._counter = 0  # 包头计数器
        self.state = ANALYSIS_STATUS.WAIT_RECV_HEADER  # 状态
        self._recvDataLength = 0  # 获取数据长度
        self.RecvPacketList = []  # 已经接收包列表

    def recv(self, buffer: bytes):
        self._buff += buffer
        self._getData()

    def _getData(self):
        while True:
            if self._buffOffset >= len(self._buff):
                break
            m_buff = self._buff[self._buffOffset:len(self._buff)]
            try:
                m_offset = m_buff.index(b'\r')
            except Exception as e:
                m_offset = 0
            if m_offset <= 0:
                break
            else:
                if self._counter == 0:
                    self.state= ANALYSIS_STATUS.WAIT_RECV_HEADER
                    self._protocol.ver = bytes.decode(
                        m_buff[0:m_offset], encoding='utf-8')
                elif self._counter == 1:
                    #cc= bytes.decode(m_buff[0:m_offset])
                    self._protocol.index = int.from_bytes(m_buff[0:m_offset],byteorder="little",signed=False)
                elif self._counter == 2:
                    self._protocol.charset = bytes.decode(
                        m_buff[0:m_offset], encoding='utf-8')
                elif self._counter == 3:
                    self._protocol.dataLength = int.from_bytes(m_buff[0:m_offset],byteorder='little',signed=False)
                elif self._counter == 4:
                    self._protocol.dataMd5 = bytes.decode(
                        m_buff[0:m_offset], encoding='utf-8')
                    self.state = ANALYSIS_STATUS.GET_DATA
                elif self._counter ==5:
                    self._protocol.command= bytes.decode(
                       m_buff[0:m_offset],encoding="utf-8" 
                    )
                elif self._counter == 6:
                    self._protocol.data += m_buff[0:len(m_buff)-1]
                    cc=len(self._protocol.data)
                    aa= self._protocol.dataLength
                    if self._protocol.dataLength == len(self._protocol.data):
                        self.state = ANALYSIS_STATUS.RECV_COMPLETE
                self._buffOffset += m_offset+1
                self._counter += 1
                if self.state == ANALYSIS_STATUS.RECV_COMPLETE:
                    self.__getDataComplete()
                    break

    def __getDataComplete(self):
        self.RecvPacketList.append(copy.deepcopy(self._protocol))
        self._protocol.clear()
        self._recvDataLength = 0
        self._buff = self._buff[self._buffOffset:len(self._buff)]
        self._buffOffset = 0
        self._counter = 0
        #self.state = ANALYSIS_STATUS.WAIT_RECV_HEADER

    def getRecvData(self):
        m_ret = []
        while True:
            m_ret.append(self.RecvPacketList[0])
            del(self.RecvPacketList[0])
            if len(self.RecvPacketList) == 0:
                break
        return m_ret

    def resetState(self):
        self.state = ANALYSIS_STATUS.WAIT_RECV_HEADER

    def getSendBuffer(self):
        return self._protocol.decode()


class ANALYSIS_STATUS(Enum):
    ERROR = 0
    WAIT_RECV_HEADER = 1
    GET_DATA = 2
    RECV_COMPLETE = 3



import datetime
import time
import shelve
import time
import threading
import asyncio
from TDhelper.Decorators.log import logging_setup, SYS_LOGS, logging
from TDhelper.Scheduler.log_config import logging_config
from TDhelper.bin.globalvar import *
from TDhelper.reflect import *
from TDhelper.Scheduler.base import *
from TDhelper.generic.requier import R

class SchedulerService(threading.Thread):
    def __init__(
        self, serializeFile="/data/", report_dir="/data/report", logging_cfg=None
    ):
        threading.Thread.__init__(self)
        logging_setup(logging_cfg if logging_cfg != None else logging_config)
        self.loop = asyncio.get_event_loop()
        self._report_dir = report_dir
        self._cacheFile = serializeFile
        self._eventLock = threading.Event()
        self._eventLock.set()
        self._timeArray = [
            ['item' for _ in range(5)] for _ in range(86400)
        ]  # 时间轮，一天的86400秒(后续可修改多重时间轮:时，分，秒，毫秒)
        self._runtasks = []  # 正在运行的任务列表
        self._index = 0  # 当前索引秒
        self._last_index = 0  # 上一次索引秒
        self.run()  # 启动任务线程
        # threading.Thread(target=self.Scheduler).start()  #启动任务列表线程

    def LoadConfig(self):
        """Load scheduler config"""
        pass

    def Install(self, task: Scheduler):
        """Install task"""
        config = task.Config()
        """config params
        config["type"]
        config["starttime"]
        config["sleep"]
        config["plug"]
        config["name"]"""

        with shelve.open(
            self._cacheFile, flag="c", writeback=True
        ) as cFile:  # open config cache file
            try:
                if config["type"].lower() == "realtime":  # run on now
                    pass
                elif config["type"].lower() == "interval":  # loop run task
                    pass
                elif config["type"].lower() == "timming":  # run on datetime with oncee
                    pass
            finally:
                cFile.close()

    def Uninstall(self, task):
        """Uninstall task"""
        pass

    async def task_list(self, timestamp):
        logging.info("async task_list TIMESTAMP(%d)." % timestamp)
        if 0 < timestamp < len(self._timeArray):
            _tasks = self._timeArray[timestamp]
            for item in _tasks:
                dynamic_cls=R("TDhelper.Scheduler.example_scheduler")
                dynamic_cls.Instance("example_scheduler",{
                    "name":"a",
                    "_plugin":"plugin",
                    "taskType":"type",
                    "startTime":datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                    "sleep":10
                })
                state, result= await dynamic_cls.call_async("run")
                if state:
                    logging.info("example result:%s",result['result'])
                else:
                    logging.error("%s",result['msg'])
        else:
            logging.error("timestamp(%d) out of range. scope(0-86400)" % timestamp)

    async def generate_report(self, datetime):
        """
        gennerate today task report.
        args:
            datetime: YYYY-MM-DD.report
        """
        pass

    def run(self, *args, **kw):
        while True:
            self._index = (
                time.localtime().tm_hour * 60 * 60
                + time.localtime().tm_min * 60
                + time.localtime().tm_sec
            )  # 当前秒索引
            if self._index != self._last_index:
                if self._index >= 86400:  # 当日计时完毕
                    logging.info("today tasks complete. task report:*.report")
                    self.loop.run_until_complete(
                        asyncio.wait([self.generate_report(datetime())])
                    )
                    break
                else:
                    tasks = [asyncio.ensure_future(self.task_list(self._index))]
                    self.loop.run_until_complete(asyncio.wait(tasks))
            time.sleep(0.5)
            self._last_index = self._index
        self.loop.close()



from enum import Enum
import functools
import threading
import TDhelper.bin.globalvar

class Event:
    '''
        Initialization event class,and set this class state flag,if has EventList Variable state is True,else state is False
    '''
    def __init__(self):
        self._state=True
        self._currentEvent= None
        if TDhelper.bin.globalvar.getGlobalVariable("EventList")==None: #Check eventList object is define
            self._state=False

    def _keyconvert(self,key):
        if isinstance(key,str):
            return (str(self.__hash__())+key).lower()
        elif isinstance(key,Enum):
            return (str(self.__hash__())+key.value).lower()
        else:
            return None
    '''
        Check event key haven't set
        Return:
        if event key haven't set return True,else it return False
    '''
    def _checkKeyNotInEventList(self,key):
        return self._keyconvert(key) not in TDhelper.bin.globalvar.getGlobalVariable("EventList")

    def _getEventListObject(self,key):
        return TDhelper.bin.globalvar.getGlobalVariable("EventList")[self._keyconvert(key)]

    def getEventList(self):
        return TDhelper.bin.globalvar.getGlobalVariable("EventList")

    '''
        Register an event
        Parameter
            key:
                event index key
            func:
                if sys has happend an event,then callback func
            sync:
                default value is True so callback by synchronous,if you set value is False then call back by asynchronous
    '''
    def registerEvent(self,key,func,sync=True):
        if self._state:
            if self._checkKeyNotInEventList(self._keyconvert(key)):
                TDhelper.bin.globalvar.getGlobalVariable("EventList")[self._keyconvert(key)]={"func":func,"sync":sync}


    '''
        Delete an event in the Event List;if you delete then this event has happend we don't callback
    '''
    def delEvent(self,key):
        if self._state:
            if not self._checkKeyNotInEventList(self._keyconvert(key)):
                del TDhelper.bin.globalvar.getGlobalVariable("EventList")[self._keyconvert(key)]

    '''
        Callback
    '''
    def on(self,key,*args,**kw):
        try:
            if key:
                if self._state:
                    if self._keyconvert(key) in TDhelper.bin.globalvar.getGlobalVariable("EventList"):
                        self._currentEvent= key
                        if self._getEventListObject(key)["sync"]:
                            return self._getEventListObject(key)["func"](*args,**kw)
                        else:
                            #todo asynchronous call back
                            pass#threading.Thread(target=self._getEventListObject(key)["func"],args=[*args,**kw,]).start()
        except Exception as e:
            raise e
    def EVENT(self):
        return self._currentEvent

def keyconvert(key):
    if isinstance(key,str):
        return key.lower()
    elif isinstance(key,Enum):
        return key.value.lower()

def call(key,*args,**kw):
    try:
        key=keyconvert(key)
        if len(args)>0:
            if key in TDhelper.bin.globalvar.getGlobalVariable("EventList"):
                if TDhelper.bin.globalvar.getGlobalVariable("EventList")[key]["sync"]:
                    return TDhelper.bin.globalvar.getGlobalVariable("EventList")[key]["func"](*args,**kw)
                else:
                    pass#threading.Thread(target=bin.globalvar.getGlobalVariable("EventList")[key]["func"],args=[*args,**kw,]).start()
    except Exception as e:
        raise e

def trigger(EventType):
        '''trigger cls function
            Description: class must inherit Event, trigger begin,complete event. 
                event begin, will given all trigger func params to the begin callback func.
                event complete,  the trigger func return will insert into kwargs with key 'result', in the callback func you can use kwargs['result'] to read trigger func return. 
            
            Params:
                EventType: <str|enum>
                
            Return:
                trigger func return.
        '''
        def decorator(func):
            @functools.wraps(func)
            def expand(*args,**kw):
                #to add event start event
                args[0].on("begin"+str(EventType),*args,**kw)
                #to add event start event
                result=func(*args,**kw)
                #to add event complate event
                kw['result']=result
                args[0].on("on"+str(EventType)+"complete",*args,**kw)
                #to add event complate event
                return result
            return expand
        return decorator 


from collections.abc import Callable
from typing import Any
class oProperty(property):
    k= ''
    ins= None
    def __init__(self,k,ins, fget: Callable[[Any], Any] | None = None, fset: Callable[[Any, Any], None] | None = None, fdel: Callable[[Any], None] | None = None, doc: str | None = None) -> None:
        self.k= k
        self.ins= ins
        fget= fget if fget else self.getter
        fset= fset if fset else self.setter
        fdel= fdel if fdel else self.deleter
        doc= doc if doc else k
        super().__init__(fget, fset, fdel, doc)
        
    def setter(self,__instance, __value) -> property:
        if self.k in __instance.__fields__:
            __instance.__fields__[self.k]= __value
        else:
            raise ValueError("do not have attribute %s"%self.k)
            
    def getter(self, __instance) -> property:
        if self.k in __instance.__fields__:
            return __instance.__fields__[self.k]
        else:
            return None
    
    def deleter(self, __fdel: Callable[[Any], None]) -> property:
        return super().deleter(__fdel)